using System;
using System.Collections.Generic;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;

namespace MapTrimmer
{
	internal unsafe class Colour
	{
		private int _r = 0;
		private int _g = 0;
		private int _b = 0;
		private int _tolerance = 5000;
		private byte _n = 0;

		internal Colour( byte r, byte g, byte b, int tolerance )
		{
			_r = r;
			_g = g;
			_b = b;
			_tolerance = tolerance;
		}

		internal Colour()
		{
		}

		internal bool Matches( byte* imagePointer )
		{
			int _bp = imagePointer[0] - _b;
			int _gp = imagePointer[1] - _g;
			int _rp = imagePointer[2] - _r;
			return ( ( _bp * _bp + _gp * _gp + _rp * _rp ) < _tolerance );
		}

		internal void TestAndAdd( byte* imagePointer, ColourList colours )
		{
			if ( !colours.Contains( imagePointer ) )
			{
				_b += imagePointer[0];
				_g += imagePointer[1];
				_r += imagePointer[2];
				_n++;
			}
		}

		internal void StoreAverage( byte* imagePointer )
		{
			if ( _n > 0 )
			{
				imagePointer[0] = (byte)( _b / _n );
				imagePointer[1] = (byte)( _g / _n );
				imagePointer[2] = (byte)( _r / _n );
			}
		}
	}

	internal unsafe class ColourList : List<Colour>
	{
		public bool Contains( byte* imagePointer )
		{
			bool retval = false;
			foreach ( Colour c in this )
			{
				if ( c.Matches( imagePointer ) )
				{
					retval = true;
					break;
				}
			}
			return retval;
		}
	}

	class Program
	{
		private static Bitmap Trim( Bitmap map, int left, int top, int right, int bottom )
		{
			Rectangle srcRect = new Rectangle( left, top, map.Width - right - left, map.Height - bottom - top );

			Rectangle destRect = new Rectangle( 0, 0, map.Width, map.Height );
			Bitmap dest = new Bitmap( map.Width, map.Height, PixelFormat.Format24bppRgb );

			Graphics g = Graphics.FromImage( dest );
			g.DrawImage( map, destRect, srcRect, GraphicsUnit.Pixel );
			return dest;
		}

		private unsafe static Bitmap RemoveColours( Bitmap map, ColourList colours )
		{
			int W = map.Width;
			int H = map.Height;

			Rectangle R = new Rectangle( 0, 0, W, H );
			BitmapData bd = map.LockBits( R, ImageLockMode.ReadWrite, PixelFormat.Format24bppRgb );
			byte* pptr = (byte*)bd.Scan0.ToPointer();
			int offset = bd.Stride - W * 3;
			for ( int y = 0; y < H; y++ )
			{
				for ( int x = 0; x < W; x++ )
				{
					if ( colours.Contains( pptr ) )
					{
						Colour newColour = new Colour();
						if ( x > 0 )
						{
							if ( y > 0 )
								newColour.TestAndAdd( pptr - bd.Stride - 3, colours );
							if ( y < ( H - 1 ) )
								newColour.TestAndAdd( pptr + bd.Stride - 3, colours );
							newColour.TestAndAdd( pptr - 3, colours );
						}

						if ( x < ( map.Width - 1 ) )
						{
							if ( y > 0 )
								newColour.TestAndAdd( pptr - bd.Stride + 3, colours );
							if ( y < ( map.Height - 1 ) )
								newColour.TestAndAdd( pptr + bd.Stride + 3, colours );
							newColour.TestAndAdd( pptr + 3, colours );
						}

						if ( y > 0 )
							newColour.TestAndAdd( pptr - bd.Stride, colours );
						if ( y < ( map.Height - 1 ) )
							newColour.TestAndAdd( pptr + bd.Stride, colours );
						newColour.StoreAverage( pptr );
					}
					pptr += 3;
				}
				pptr += offset;
			}
			map.UnlockBits( bd );
			return map;
		}

		private const string defaultInput = @"C:\Program Files (x86)\Bethesda Softworks\Oblivion\Data\Textures\Maps\Cvt";
		private const string defaultOutput = @"C:\Program Files (x86)\Bethesda Softworks\Oblivion\Data\Textures\Maps\Done\";

		static void Main( string[] args )
		{
			try
			{
				bool gotInput = false;
				string inputPath = string.Empty;
				while ( !gotInput )
				{
					Console.WriteLine( "Please enter the full pathname of the directory with the input files" );
					Console.WriteLine( "(enter a blank string to use the default {0})", defaultInput );
					inputPath = Console.ReadLine();
					if ( inputPath.Length == 0 )
						inputPath = defaultInput;
					gotInput = Directory.Exists( inputPath );
				}

				Console.WriteLine( "Please enter the full path where the new files should be stored" );
				Console.WriteLine( "(enter a blank string to use the default {0})", defaultOutput );
				string outputPath = Console.ReadLine();
				if ( outputPath.Length == 0 )
					outputPath = defaultOutput;
				Directory.CreateDirectory( outputPath );
				if ( !outputPath.EndsWith( Path.DirectorySeparatorChar.ToString() ) )
					outputPath += Path.DirectorySeparatorChar;

				ColourList colours = new ColourList();
				colours.Add( new Colour( 0xff, 0xff, 0x00, 500 ) ); // pure yellow
				colours.Add( new Colour( 0xff, 0xff, 0x3d, 500 ) ); // yellow
				colours.Add( new Colour( 0x45, 0xff, 0x00, 500 ) ); // green
				colours.Add( new Colour( 0x00, 0xff, 0x40, 5000 ) ); // greenish
				colours.Add( new Colour( 0xff, 0x00, 0x00, 5000 ) ); // pure red
				colours.Add( new Colour( 0x90, 0xaa, 0xd0, 5000 ) ); // blueish
				colours.Add( new Colour( 0xcc, 0x30, 0x30, 5000 ) ); // reddish

				DirectoryInfo di = new DirectoryInfo( inputPath );
				FileInfo[] inputMaps = di.GetFiles( "*.bmp" );
				foreach ( FileInfo fi in inputMaps )
				{
					Console.WriteLine( fi.Name );
					Bitmap oneMap = new Bitmap( fi.FullName );
					oneMap = RemoveColours( oneMap, colours );
					oneMap = Trim( oneMap, 0, 14, 15, 1 );
					oneMap.Save( outputPath + fi.Name, ImageFormat.Bmp );
				}
			}
			catch ( Exception catchAll )
			{
				Console.WriteLine( "An error occurred: {0}", catchAll.Message );
			}
		}
	}
}